#
##
##  SPDX-FileCopyrightText: © 2007-2021 Benedict Verhegghe <bverheg@gmail.com>
##  SPDX-License-Identifier: GPL-3.0-or-later
##
##  This file is part of pyFormex 2.6  (Mon Aug 23 15:13:50 CEST 2021)
##  pyFormex is a tool for generating, manipulating and transforming 3D
##  geometrical models by sequences of mathematical operations.
##  Home page: https://pyformex.org
##  Project page: https://savannah.nongnu.org/projects/pyformex/
##  Development: https://gitlab.com/bverheg/pyformex
##  Distributed under the GNU General Public License version 3 or later.
##
##  This program is free software: you can redistribute it and/or modify
##  it under the terms of the GNU General Public License as published by
##  the Free Software Foundation, either version 3 of the License, or
##  (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##  GNU General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see http://www.gnu.org/licenses/.
##
#
"""Setup script for pyFormex

To change the install location, see the prefix and exec-prefix in setup.cfg.
To uninstall pyFormex: pyformex --remove
"""
import sys

# Detect platform
if sys.version_info < (3, 6):
    raise RuntimeError("Python version is %s.%s but pyFormex requires "
                       ">= 3.6." % sys.version_info[:2])

import numpy as np
from setuptools import setup, Extension

# pyFormex release
__RELEASE__ = '2.6'

# define the things to include
import manifest

# The acceleration libraries
LIB_MODULES = [ 'misc_', 'nurbs_']

ext_modules = [
    Extension(f"pyformex.lib.{m}", sources = [f"pyformex/lib/{m}.c"],
              include_dirs=[np.get_include()]) for m in LIB_MODULES
]

ext_modules.append(Extension("pyformex.lib._clustering",
                             ["pyformex/lib/_clustering.pyx"],
                             language='c++'))

def print_msgs(msgs):
    """Print status messages"""
    print('*' * 75)
    for msg in msgs:
        print(msg)
    print('*' * 75)


def run_setup(accel):
    kargs = {}
    if accel:
        kargs['ext_modules'] = ext_modules

    with open('Description') as file:
        long_description = file.read()

    # PKG_DATA, relative from pyformex path
    PKG_DATA = [
        'pyformexrc',
        'icons/README',
        'icons/*.xpm',
        'icons/*.gif',
        'icons/pyformex*.png',
        'icons/64x64/*',
        'fonts/*',
        'glsl/*',
        'examples/apps.cat',
        'bin/*',
        'data/*',
        'doc/*',
        'sphinx/**',
        'extra/Makefile',
        'extra/*/*',
        'scripts/*',
        ]

    setup(
        #use_scm_version=True,
        #setup_requires=['setuptools_scm'],
        name='pyformex',
        version=__RELEASE__,
        description='Create, transform, manipulate, render 3D geometry from Python scripts',
        long_description=long_description,
        author='Benedict Verhegghe',
        author_email='bverheg@gmail.com',
        url='http://pyformex.org',
        download_url='http://download.savannah.gnu.org/releases/pyformex/pyformex-%s.tar.gz' % __RELEASE__,
        license='GNU General Public License (GPL)',
        packages=[
            'pyformex',
            'pyformex.gui',
            'pyformex.gui.menus',
            'pyformex.lib',
            'pyformex.opengl',
            'pyformex.plugins',
            'pyformex.appsdir',
            #  'pyformex.scripts',  # this is not a package!
            'pyformex.examples',
            'pyformex.fe',
            'pyformex.freetype',
            'pyformex.freetype.ft_enums',
            'pyformex.vtk_light',
            'pyformex.vtk_light.util',
        ],
        include_package_data=True,
        #package_data={ 'pyformex': PKG_DATA },
        scripts=['pyformex/pyformex'],
        data_files=manifest.OTHER_DATA,
        classifiers=[
            'Development Status :: 4 - Beta',
            'Environment :: Console',
            'Environment :: X11 Applications :: Qt',
            'Intended Audience :: End Users/Desktop',
            'Intended Audience :: Science/Research',
            'Intended Audience :: Education',
            'License :: OSI Approved :: GNU General Public License (GPL)',
            'Operating System :: POSIX :: Linux',
            'Operating System :: POSIX',
            'Operating System :: OS Independent',
            'Programming Language :: Python',
            'Programming Language :: C',
            'Topic :: Multimedia :: Graphics :: 3D Modeling',
            'Topic :: Multimedia :: Graphics :: 3D Rendering',
            'Topic :: Scientific/Engineering :: Mathematics',
            'Topic :: Scientific/Engineering :: Visualization',
            'Topic :: Scientific/Engineering :: Physics',
        ],
        #          requires=['numpy','OpenGL','PyQt4 | PySide'],
        **kargs
    )


# Detect some --no-??? options
options = {
    'accel': True,
    'cython': True,
    'sphinx': True,
    }
for opt in options:
    try:
        i = sys.argv.index(f'--no-{opt}')
        del(sys.argv[i])
        options[opt] = False
    except ValueError:
        pass

msgs = []

if options['cython']:
    try:
        import cython
    except ImportError:
        msgs.append("""\
Cython is missing, so I can't compile the clustering extension.
You should install the package cython3.
Or you can setup pyFormex without the extension, using the command:
    python setup.py build --no-cython""")

if options['sphinx']:
    try:
        import sphinx
    except ImportError:
        msgs.append("""\
Sphinx is missing, so I can't compile the local html documentation.
You should install the package python3-sphinx.
Or you can setup pyFormex without local docs, using the command:
    python setup.py build --no-sphinx""")

if msgs:
    print_msgs(msgs)
    sys.exit()

try:
    run_setup(options)
except Exception as e:
    msgs.append(e)
    if options['accel']:
        msgs.append("""\
If the installation failed because the acceleration libraries could
not be compiled, you may try installing without them:
    python setup.py build --no-accel""")

if msgs:
    print_msgs(msgs)

# End
